# keyword.py - $Keyword$ expansion for Mercurial
#
# Copyright 2007 Christian Ebert <blacktrash@gmx.net>
#
# This software may be used and distributed according to the terms
# of the GNU General Public License, incorporated herein by reference.
#
# $Id$
#
# Keyword expansion hack against the grain of a DSCM
#
# There are many good reasons why this is not needed in a distributed
# SCM, still it may be useful in very small projects based on single
# files (like LaTeX packages), that are mostly addressed to an audience
# not running a version control system.
#
# For in-depth discussion refer to
# <http://www.selenic.com/mercurial/wiki/index.cgi/KeywordPlan>.
#
# Keyword expansion is based on Mercurial's changeset template mappings.
#
# Binary files are not touched.
#
# Setup in hgrc:
#
#   [extensions]
#   # enable extension
#   keyword = /full/path/to/hgkw/keyword.py
#   # or, if script in canonical hgext folder:
#   # hgext.keyword =
#
# Files to act upon/ignore are specified in the [keyword] section.
# Customized keyword template mappings in the [keywordmaps] section.
#
# Run "hg help keyword" and "hg kwdemo" to get info on configuration.

'''keyword expansion in local repositories

This extension expands RCS/CVS-like or self-customized $Keywords$
in tracked text files selected by your configuration.

Keywords are only expanded in local repositories and not stored in
the change history. The mechanism can be regarded as a convenience
for the current user or for archive distribution.

Configuration is done in the [keyword] and [keywordmaps] sections
of hgrc files.

Example:

    [keyword]
    # expand keywords in every python file except those matching "x*"
    **.py =
    x*    = ignore

Note: the more specific you are in your filename patterns
      the less you lose speed in huge repos.

For [keywordmaps] template mapping and expansion demonstration and
control run "hg kwdemo".

An additional date template filter {date|utcdate} is provided.

The default template mappings (view with "hg kwdemo -d") can be replaced
with customized keywords and templates.
Again, run "hg kwdemo" to control the results of your config changes.

Before changing/disabling active keywords, run "hg kwshrink" to avoid
the risk of inadvertedly storing expanded keywords in the change history.

To force expansion after enabling it, or a configuration change, run
"hg kwexpand".

Expansions spanning more than one line and incremental expansions,
like CVS' $Log$, are not supported. A keyword template map
"Log = {desc}" expands to the first line of the changeset description.

Caveat: "hg import" fails if the patch context contains an active
        keyword. In that case run "hg kwshrink", and then reimport.
        Or, better, use bundle/unbundle to share changes.
'''

from mercurial import commands, cmdutil, context, fancyopts
from mercurial import filelog, localrepo, revlog, templater, util
from mercurial.i18n import gettext as _
import getopt, os.path, re, shutil, sys, tempfile, time

# backwards compatibility hacks

try:
    # cmdutil.parse moves to dispatch._parse in 18a9fbb5cd78
    from mercurial import dispatch
    _parse = dispatch._parse
except ImportError:
    try:
        # commands.parse moves to cmdutil.parse in 0c61124ad877
        _parse = cmdutil.parse
    except AttributeError:
        _parse = commands.parse

try:
    # bail_if_changed moves from commands to cmdutil in 0c61124ad877
    bail_if_changed = cmdutil.bail_if_changed
except AttributeError:
    bail_if_changed = commands.bail_if_changed

def _pathto(repo, f, cwd=None):
    '''kwfiles behaves similar to status, using pathto since 78b6add1f966.'''
    try:
        if cwd is None:
            cwd = repo.getcwd()
        return repo.pathto(f, cwd)
    except AttributeError:
        return f

# commands.parse/cmdutil.parse returned nothing for
# "hg diff --rev" before 88803a69b24a due to bug in fancyopts
def _fancyopts(args, options, state):
    '''Fixed fancyopts from 88803a69b24a.'''
    long = []
    short = ''
    map = {}
    dt = {}
    for s, l, d, c in options:
        pl = l.replace('-', '_')
        map['-'+s] = map['--'+l] = pl
        if isinstance(d, list):
            state[pl] = d[:]
        else:
            state[pl] = d
        dt[pl] = type(d)
        if (d is not None and d is not True and d is not False and
            not callable(d)):
            if s: s += ':'
            if l: l += '='
        if s: short = short + s
        if l: long.append(l)
    opts, args = getopt.getopt(args, short, long)
    for opt, arg in opts:
        if dt[map[opt]] is type(fancyopts): state[map[opt]](state, map[opt], arg)
        elif dt[map[opt]] is type(1): state[map[opt]] = int(arg)
        elif dt[map[opt]] is type(''): state[map[opt]] = arg
        elif dt[map[opt]] is type([]): state[map[opt]].append(arg)
        elif dt[map[opt]] is type(None): state[map[opt]] = True
        elif dt[map[opt]] is type(False): state[map[opt]] = True
    return args

fancyopts.fancyopts = _fancyopts


commands.optionalrepo += ' kwdemo'

def utcdate(date):
    '''Returns hgdate in cvs-like UTC format.'''
    return time.strftime('%Y/%m/%d %H:%M:%S', time.gmtime(date[0]))

class kwtemplater(object):
    '''
    Sets up keyword templates, corresponding keyword regex, and
    provides keyword substitution functions.
    '''
    templates = {
        'Revision': '{node|short}',
        'Author': '{author|user}',
        'Date': '{date|utcdate}',
        'RCSFile': '{file|basename},v',
        'Source': '{root}/{file},v',
        'Id': '{file|basename},v {node|short} {date|utcdate} {author|user}',
        'Header': '{root}/{file},v {node|short} {date|utcdate} {author|user}',
    }

    def __init__(self, ui, repo, expand, path='', node=None):
        self.ui = ui
        self.repo = repo
        self.t = expand or None
        self.path = path
        self.node = node
        self.debug = self.ui.debugflag

        kwmaps = self.ui.configitems('keywordmaps')
        if kwmaps: # override default templates
            kwmaps = [(k, templater.parsestring(v, quoted=False))
                      for (k, v) in kwmaps]
            self.templates = dict(kwmaps)
        escaped = map(re.escape, self.templates.keys())
        kwpat = r'\$(%s)(: [^$\n\r]*? )??\$' % '|'.join(escaped)
        self.re_kw = re.compile(kwpat)
        if self.t:
            templater.common_filters['utcdate'] = utcdate
            self.t = self._changeset_templater()

    def _changeset_templater(self):
        '''Backwards compatible cmdutil.changeset_templater.'''
        # before 1e0b94cfba0e there was an extra "brinfo" argument
        try:
            return cmdutil.changeset_templater(self.ui, self.repo,
                                               False, '', False)
        except TypeError:
            return cmdutil.changeset_templater(self.ui, self.repo,
                                               False, None, '', False)

    def _wwrite(self, f, data, man):
        '''Makes repo.wwrite backwards compatible.'''
        # 656e06eebda7 removed file descriptor argument
        # 67982d3ee76c added flags argument
        try:
            self.repo.wwrite(f, data, man.flags(f))
        except (AttributeError, TypeError):
            self.repo.wwrite(f, data)

    def _normal(self, files):
        '''Backwards compatible repo.dirstate.normal/update.'''
        # 6fd953d5faea introduced dirstate.normal()
        try:
            for f in files:
                self.repo.dirstate.normal(f)
        except AttributeError:
            self.repo.dirstate.update(files, 'n')

    def kwsub(self, mobj):
        '''Substitutes keyword using corresponding template.'''
        kw = mobj.group(1)
        self.t.use_template(self.templates[kw])
        self.ui.pushbuffer()
        self.t.show(changenode=self.node, root=self.repo.root, file=self.path)
        keywordsub = templater.firstline(self.ui.popbuffer())
        return '$%s: %s $' % (kw, keywordsub)

    def substitute(self, node, data, subfunc):
        '''Obtains node if missing.
        Ensures consistent templates regardless of ui.debugflag.
        Calls given substitution function.'''
        if not self.node:
            c = context.filectx(self.repo, self.path, fileid=node)
            self.node = c.node()
        self.ui.debugflag = False
        result = subfunc(self.kwsub, data)
        self.ui.debugflag = self.debug
        return result

    def expand(self, node, data):
        '''Returns data with keywords expanded.'''
        if util.binary(data):
            return data
        return self.substitute(node, data, self.re_kw.sub)

    def process(self, node, data):
        '''Returns a tuple: data, count.
        Count is number of keywords/keyword substitutions.
        Keywords in data are expanded, if templater was initialized.'''
        if util.binary(data):
            return data, None
        if self.t:
            return self.substitute(node, data, self.re_kw.subn)
        return data, self.re_kw.search(data)

    def shrink(self, text):
        '''Returns text with all keyword substitutions removed.'''
        if util.binary(text):
            return text
        return self.re_kw.sub(r'$\1$', text)

    def overwrite(self, candidates, man, commit):
        '''Overwrites files in working directory if keywords are detected.
        Keywords are expanded if keyword templater is initialized,
        otherwise their substitution is removed.'''
        expand = self.t is not None
        action = ('shrinking', 'expanding')[expand]
        notify = (self.ui.note, self.ui.debug)[commit]
        overwritten = []
        for f in candidates:
            fp = self.repo.file(f, kwexp=expand, kwmatch=True)
            data, kwfound = fp.kwctread(man[f])
            if kwfound:
                notify(_('overwriting %s %s keywords\n') % (f, action))
                self._wwrite(f, data, man)
                overwritten.append(f)
        self._normal(overwritten)

class kwfilelog(filelog.filelog):
    '''
    Subclass of filelog to hook into its read, add, cmp methods.
    Keywords are "stored" unexpanded, and processed on reading.
    '''
    def __init__(self, opener, path, kwtemplater):
        super(kwfilelog, self).__init__(opener, path)
        self.kwtemplater = kwtemplater

    def kwctread(self, node):
        '''Reads expanding and counting keywords
        (only called from kwtemplater.overwrite).'''
        data = super(kwfilelog, self).read(node)
        return self.kwtemplater.process(node, data)

    def read(self, node):
        '''Expands keywords when reading filelog.'''
        data = super(kwfilelog, self).read(node)
        return self.kwtemplater.expand(node, data)

    def add(self, text, meta, tr, link, p1=None, p2=None):
        '''Removes keyword substitutions when adding to filelog.'''
        text = self.kwtemplater.shrink(text)
        return super(kwfilelog, self).add(text, meta, tr, link, p1=p1, p2=p2)

    def cmp(self, node, text):
        '''Removes keyword substitutions for comparison.'''
        text = self.kwtemplater.shrink(text)
        if self.renamed(node):
            t2 = super(kwfilelog, self).read(node)
            return t2 != text
        return revlog.revlog.cmp(self, node, text)

def _bail_if_nokwconf(ui):
    if hasattr(ui, 'kwfmatcher'):
        return
    if ui.configitems('keyword'):
        raise util.Abort(_('[keyword] patterns cannot match'))
    raise util.Abort(_('no [keyword] patterns configured'))

def _iskwfile(ui, man, f):
    return not man.linkf(f) and ui.kwfmatcher(f)

def _overwrite(ui, repo, files, node, man, expand, commit):
    '''Passes given files to kwtemplater for overwriting.'''
    if files:
        files.sort()
        kwt = kwtemplater(ui, repo, expand, node=node)
        kwt.overwrite(files, man, commit)

def _kwfwrite(ui, repo, expand, *pats, **opts):
    '''Selects files and passes them to _overwrite.'''
    _bail_if_nokwconf(ui)
    bail_if_changed(repo)
    wlock = lock = None
    try:
        wlock = repo.wlock()
        lock = repo.lock()
        files, match, anypats = cmdutil.matchpats(repo, pats, opts)
        fdict = dict.fromkeys(files)
        fdict.pop('.', None)
        ctx = repo.changectx()
        man = ctx.manifest()
        mfiles = man.keys()
        mfiles.sort()
        files = []
        for f in mfiles:
            for ff in fdict:
                if ff == f or ff.startswith('%s/' % f):
                    if _iskwfile(ui, man, ff):
                        files.append(ff)
                    del fdict[ff]
                    break
            if not f in files and match(f) and _iskwfile(ui, man, f):
                files.append(f)
        ffiles = fdict.keys()
        ffiles.sort()
        for f in ffiles:
            ui.warn(_('%s: No such file in working copy\n') % _pathto(repo, f))
        # 7th argument sets commit to False
        _overwrite(ui, repo, files, ctx.node(), man, expand, False)
    finally:
        del wlock, lock


def shrink(ui, repo, *pats, **opts):
    '''revert expanded keywords in working directory

    Run before changing/disabling active keywords
    or if you experience problems with "hg import" or "hg merge".
    '''
    # 3rd argument sets expansion to False
    _kwfwrite(ui, repo, False, *pats, **opts)

def expand(ui, repo, *pats, **opts):
    '''expand keywords in working directory

    Run after (re)enabling keyword expansion.
    '''
    # 3rd argument sets expansion to True
    _kwfwrite(ui, repo, True, *pats, **opts)

def files(ui, repo, *pats, **opts):
    '''print files currently configured for keyword expansion

    Crosscheck which files in working directory are potential targets for
    keyword expansion.
    That is, files matched by [keyword] config patterns but not symlinks.
    '''
    _bail_if_nokwconf(ui)
    files, match, anypats = cmdutil.matchpats(repo, pats, opts)
    status = repo.status(files=files, match=match, list_clean=True)
    modified, added, removed, deleted, unknown, ignored, clean = status
    if opts['untracked']:
        files = modified + added + unknown + clean
    else:
        files = modified + added + clean
    files.sort()
    # use the full definition of repo._link for backwards compatibility
    kwfiles = [f for f in files if ui.kwfmatcher(f)
               and not os.path.islink(repo.wjoin(f))]
    cwd = pats and repo.getcwd() or ''
    allf = opts['all']
    ignore = opts['ignore']
    flag = (allf or ui.verbose) and 1 or 0
    if not ignore:
        format = ('%s\n', 'K %s\n')[flag]
        for k in kwfiles:
            ui.write(format % _pathto(repo, k, cwd))
    if allf or ignore:
        format = ('%s\n', 'I %s\n')[flag]
        for i in [f for f in files if f not in kwfiles]:
            ui.write(format % _pathto(repo, i, cwd))

def demo(ui, repo, *args, **opts):
    '''print [keywordmaps] configuration and an expansion example

    Show current, custom, or default keyword template maps
    and their expansion.

    Extend current configuration by specifying maps as arguments
    and optionally by reading from an additional hgrc file.

    Override current keyword template maps with "default" option.
    '''
    def demostatus(stat):
        ui.status(_('\n\t%s\n') % stat)

    def demoitems(section, items):
        ui.write('[%s]\n' % section)
        for k, v in items:
            ui.write('%s = %s\n' % (k, v))

    msg = 'hg keyword config and expansion example'
    kwstatus = 'current'
    fn = 'demo.txt'
    branchname = 'demobranch'
    tmpdir = tempfile.mkdtemp('', 'kwdemo.')
    ui.note(_('creating temporary repo at %s\n') % tmpdir)
    repo = localrepo.localrepository(ui, path=tmpdir, create=True)
    ui.setconfig('keyword', fn, '')
    if args or opts['rcfile']:
        kwstatus = 'custom'
    if opts['rcfile']:
        ui.readconfig(opts['rcfile'])
    if opts['default']:
        kwstatus = 'default'
        kwmaps = kwtemplater.templates
        if ui.configitems('keywordmaps'):
            # override maps from optional rcfile
            for k, v in kwmaps.items():
                ui.setconfig('keywordmaps', k, v)
    elif args:
        # simulate hgrc parsing
        rcmaps = ['[keywordmaps]\n'] + [a + '\n' for a in args]
        fp = repo.opener('hgrc', 'w')
        fp.writelines(rcmaps)
        fp.close()
        ui.readconfig(repo.join('hgrc'))
    if not opts['default']:
        kwmaps = dict(ui.configitems('keywordmaps')) or kwtemplater.templates
    reposetup(ui, repo)
    for k, v in ui.configitems('extensions'):
        if k.endswith('keyword'):
            extension = '%s = %s' % (k, v)
            break
    demostatus('config using %s keyword template maps' % kwstatus)
    ui.write('[extensions]\n%s\n' % extension)
    demoitems('keyword', ui.configitems('keyword'))
    demoitems('keywordmaps', kwmaps.items())
    keywords = '$' + '$\n$'.join(kwmaps.keys()) + '$\n'
    repo.wopener(fn, 'w').write(keywords)
    repo.add([fn])
    path = repo.wjoin(fn)
    ui.note(_('\n%s keywords written to %s:\n') % (kwstatus, path))
    ui.note(keywords)
    ui.note('\nhg -R "%s" branch "%s"\n' % (tmpdir, branchname))
    # silence branch command if not verbose
    quiet = ui.quiet
    verbose = ui.verbose
    ui.quiet = not verbose
    commands.branch(ui, repo, branchname)
    ui.quiet = quiet
    ui.note('hg -R "%s" ci -m "%s"\n' % (tmpdir, msg))
    repo.commit(text=msg)
    pathinfo = ('', ' in %s' % path)[ui.verbose]
    demostatus('%s keywords expanded%s' % (kwstatus, pathinfo))
    ui.write(repo.wread(fn))
    ui.debug(_('\nremoving temporary repo %s\n') % tmpdir)
    shutil.rmtree(tmpdir, ignore_errors=True)


def reposetup(ui, repo):
    '''Sets up repo as kwrepo for keyword substitution.
    Overrides file method to return kwfilelog instead of filelog
    if file matches user configuration.
    Wraps commit to overwrite configured files with updated
    keyword substitutions.
    This is done for local repos only, and only if there are
    files configured at all for keyword substitution.'''

    nokwcommands = ['add', 'addremove', 'bundle', 'clone', 'copy', 'export',
                    'grep', 'identify', 'incoming', 'init', 'outgoing', 'push',
                    'remove', 'rename', 'rollback']

    if not repo.local() or _parse(ui, sys.argv[1:])[0] in nokwcommands:
        return

    inc, exc = [], ['.hgtags']
    for pat, opt in ui.configitems('keyword'):
        if opt != 'ignore':
            inc.append(pat)
        else:
            exc.append(pat)
    if not inc:
        return

    ui.kwfmatcher = util.matcher(repo.root, inc=inc, exc=exc)[1]

    class kwrepo(repo.__class__):
        def file(self, f, kwexp=True, kwmatch=False):
            if f[0] == '/':
                f = f[1:]
            if kwmatch or ui.kwfmatcher(f):
                kwt = kwtemplater(ui, self, kwexp, path=f)
                return kwfilelog(self.sopener, f, kwt)
            return filelog.filelog(self.sopener, f)

        def _commit(self, files, text, user, date, match, force, lock, wlock,
                    force_editor, p1, p2, extra):
            '''Private commit wrapper for backwards compatibility.'''
            try:
                return super(kwrepo, self).commit(files=files, text=text,
                                                  user=user, date=date,
                                                  match=match, force=force,
                                                  lock=lock, wlock=wlock,
                                                  force_editor=force_editor,
                                                  p1=p1, p2=p2, extra=extra)
            except TypeError:
                return super(kwrepo, self).commit(files=files, text=text,
                                                  user=user, date=date,
                                                  match=match, force=force,
                                                  force_editor=force_editor,
                                                  p1=p1, p2=p2, extra=extra)

        def commit(self, files=None, text='', user=None, date=None,
                   match=util.always, force=False, lock=None, wlock=None,
                   force_editor=False, p1=None, p2=None, extra={}):
            # (w)lock arguments removed in 126f527b3ba3
            # so they are None or what was passed to commit
            # use private _(w)lock for deletion
            _lock = lock
            _wlock = wlock
            del wlock, lock
            try:
                if not _wlock:
                    _wlock = self.wlock()
                if not _lock:
                    _lock = self.lock()
                node = self._commit(files, text, user, date, match, force,
                                    _lock, _wlock, force_editor, p1, p2, extra)
                if node is not None:
                    cl = self.changelog.read(node)
                    mn = self.manifest.read(cl[0])
                    candidates = [f for f in cl[3] if mn.has_key(f)
                                  and _iskwfile(ui, mn, f)]
                    # 6th, 7th arguments set expansion, commit to True
                    _overwrite(ui, self, candidates, node, mn, True, True)
                return node
            finally:
                del _wlock, _lock

    repo.__class__ = kwrepo


cmdtable = {
    'kwdemo':
        (demo,
         [('d', 'default', None, _('show default keyword template maps')),
          ('f', 'rcfile', [], _('read maps from RCFILE'))],
         _('hg kwdemo [-d] [-f RCFILE] [TEMPLATEMAP ...]')),
    'kwfiles':
        (files,
         [('a', 'all', None, _('show keyword status flags of all files')),
          ('i', 'ignore', None, _('show files excluded from expansion')),
          ('u', 'untracked', None, _('additionally show untracked files')),
         ] + commands.walkopts,
         _('hg kwfiles [OPTION]... [FILE]...')),
    'kwshrink': (shrink, commands.walkopts,
                 _('hg kwshrink [OPTION]... [FILE]...')),
    'kwexpand': (expand, commands.walkopts,
                 _('hg kwexpand [OPTION]... [FILE]...')),
}
